package ava.parser

import java.io.BufferedReader

/** Used to consume characters from the input stream.
  *
  * @param input
  *   Input stream capable of reading _characters_ rather than bytes. This
  *   stream is owned by the character reader.
  */
class CharacterReader(
  private val input: BufferedReader
):
  // Internal constants.
  private val Capacity: Int         = 4096
  private val LookBackCapacity: Int = 16

  // Internal buffers.
  private val lastChars: Ring[Char] = Ring[Char](LookBackCapacity, 0)
  private val buffer: Array[Char]   = Array.fill(Capacity)(0)

  // Internal mutable state.
  private var length: Int          = 0
  private var index: Int           = 0
  private var eof: Boolean         = false
  private var peekedAhead: Boolean = false
  private var lookAhead: Char      = 0

  /** Close the underlying stream.
    */
  def close(): Unit = input.close()

  /** Set EOF and close the underlying stream.
    */
  private def setEof(): Unit =
    eof = true
    close()

  /** Used when the buffer has been fully consumed. Reads as many characters as
    * possible and resets internal state. Detects EOF and sets state
    * accordingly.
    */
  private def fillBuffer(): Unit =
    // TODO: Short circuit on EOF!
    var numberOfCharacters = 0
    if peekedAhead then
      buffer(0) = lookAhead
      numberOfCharacters = input.read(buffer, 1, Capacity - 1)
    else numberOfCharacters = input.read(buffer, 0, Capacity)

    // Record the number of characters ACTUALLY consumed by the stream.
    // This is our working buffer size.
    length = numberOfCharacters

    // If no characters could be read, we're done.
    // EDGE CASE: If we peeked ahead to the last character, we still technically
    // have a valid buffer of length 1.
    val _ =
      if numberOfCharacters <= 0 && peekedAhead then
        // Edge case: buffer of size 1.
        length = 1
      else if length <= 0 then
        // EOF case: no characters remain.
        length = 0
        setEof()
        input.close()
      else
        // Normal case: we read some number of characters.
        length = numberOfCharacters + (if peekedAhead then 1 else 0)

    // Reset the peeked state -- we already consumed it.
    peekedAhead = false

    // Reset the index position to start iterating through the buffer again.
    index = 0

  /** Observe, but do not consume, the next character in the stream. Note that
    * if the buffer is fully consumed, this function must read the underlying
    * stream but will _not_ refresh the buffer until the next character is
    * consumed.
    *
    * @return
    *   The next character value, or None if the stream is EOF.
    */
  def peek(): Option[Char] =
    if eof then None
    else if index == length then
      // Special case -- try to read one.
      peekedAhead = true
      lookAhead = input.read().toChar
      if lookAhead < 0 then
        eof = true
        None
      else Some(lookAhead)
    else Some(buffer(index))

  def consume(): Option[Char] =
    if eof then None
    else if peekedAhead then
      fillBuffer()
      lastChars.push(lookAhead)
      Some(lookAhead)
    else if index == length then
      fillBuffer()
      if eof then None
      else
        val pos = index
        index = index + 1
        lastChars.push(buffer(pos))
        Some(buffer(pos))
    else
      val pos = index
      index = index + 1
      lastChars.push(buffer(pos))
      Some(buffer(pos))

  /** @return
    *   True if all data in the _buffer_ has been consumed, false otherwise.
    */
  def isBufferExhausted(): Boolean =
    index >= length

  /** @return
    *   True if the stream is EOF, false otherwise. An EOF stream does NOT imply
    *   that the buffer is exhausted.
    */
  def isEof(): Boolean =
    eof
